/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"
#include "http.h"
#include "digmd5.h"

/*
 * TODO: see if sequential write via mmap (read() directly mmap()-ed address)
 * improves performance.  This means we can avoid an extra pwrite() syscall
 * to the disk (essentially a reverse sendfile()).
 *
 * We'll always need to keep the pwrite() code to deal with ENOMEM
 * situations (especially on 32-bit), though :<
 */

static void file_close_null(struct mog_http *http)
{
	mog_file_close(http->forward);
	http->forward = NULL;
}

bool mog_http_write_full(struct mog_fd *file_mfd, char *buf, size_t buf_len)
{
	struct mog_file *file = &file_mfd->as.file;
	ssize_t w;

	if (file->md5)
		md5_process_bytes(buf, buf_len, file->md5);

	for (;;) {
		w = pwrite(file_mfd->fd, buf, buf_len, file->foff);

		if (w >= 0) {
			file->foff += w;
			if (w == buf_len)
				return true;

			buf_len -= w;
			buf += w;
			continue;
		}
		if (errno != EINTR)
			break;
	}

	PRESERVE_ERRNO(do {
		syslog(LOG_ERR, "pwrite() failed at offset: %lld: %m",
		       (long long)file->foff);
	} while (0));

	return false;
}

#define stop(http,status) stop0((http),(status),sizeof(status)-1);

MOG_NOINLINE static enum mog_next
stop0(struct mog_http *http, const char *status, size_t status_len)
{
	if (status) {
		struct iovec iov;
		union { const char *in; char *out; } deconst;

		deconst.in = status;
		iov.iov_base = deconst.out;
		iov.iov_len = status_len;

		mog_http_resp0(http, &iov, false);
	}
	file_close_null(http);
	return MOG_NEXT_CLOSE;
}

static bool md5_ok(struct mog_http *http)
{
	struct md5_ctx *md5 = http->forward->as.file.md5;
	uint16_t result[16];

	/* PUT requests don't _require_ Content-MD5 header/trailer */
	if (md5 == NULL)
		return true;

	md5_finish_ctx(md5, result);

	return (memcmp(http->expect_md5, result, 16) == 0);
}

static enum mog_next http_put_commit(struct mog_http *http)
{
	if (!md5_ok(http))
		return stop(http, "400 Bad Request");
	if (fchmod(http->forward->fd, http->svc->put_perms) != 0)
		return stop(http, "500 Internal Server Error");

	file_close_null(http);
	mog_http_resp(http, "201 Created", true);
	if (http->wbuf && http->wbuf != MOG_WR_ERROR)
		return MOG_NEXT_WAIT_WR;
	if (!http->persistent || http->wbuf == MOG_WR_ERROR)
		return MOG_NEXT_CLOSE;
	mog_http_reset_parser(http);
	return MOG_NEXT_ACTIVE;
}

static void commit_oneshot(struct mog_http *http)
{
	if (!md5_ok(http))
		mog_http_resp(http, "400 Bad Request", true);
	else if (fchmod(http->forward->fd, http->svc->put_perms) == 0)
		mog_http_resp(http, "201 Created", true);
	else
		mog_http_resp(http, "500 Internal Server Error", false);

	file_close_null(http);
}

static void stash_advance_rbuf(struct mog_http *http, char *buf, size_t buf_len)
{
	struct mog_rbuf *rbuf = http->rbuf;
	size_t end = http->line_end + 1;

	if (http->line_end == 0 || buf_len <= end) {
		http->offset = 0;
		mog_rbuf_free_and_null(&http->rbuf);
		return;
	}

	assert(buf[http->line_end] == '\n' && "line_end is not LF");
	assert(buf_len <= MOG_HTTP_RBUF_SIZE && "bad rbuf size");
	assert(end <= http->offset && "invalid line end");
	if (rbuf == NULL)
		http->rbuf = rbuf = mog_rbuf_new(MOG_HTTP_RBUF_SIZE);

	memmove(rbuf->rptr, buf + end, buf_len - end);
	rbuf->rsize = buf_len - end;
	http->offset -= end;
	http->line_end = 0;
}

static void
chunked_body_after_header(struct mog_http *http, char *buf, size_t buf_len)
{
	size_t tmpoff = http->offset;

	mog_chunk_init(http);
	http->offset = tmpoff;

	switch (mog_chunk_parse(http, buf, buf_len)) {
	case MOG_PARSER_ERROR:
		file_close_null(http);
		mog_http_resp(http, "400 Bad Request", false);
		return;
	case MOG_PARSER_CONTINUE:
		assert(http->chunk_state != MOG_CHUNK_STATE_DONE);
		/* fall through */
	case MOG_PARSER_DONE:
		switch (http->chunk_state) {
		case MOG_CHUNK_STATE_SIZE:
			assert(http->offset == buf_len
			       && "HTTP chunk parser didn't finish size");
			return;
		case MOG_CHUNK_STATE_DATA:
			assert(http->offset == buf_len
			       && "HTTP chunk parser didn't finish data");
			return;
		case MOG_CHUNK_STATE_TRAILER:
			assert(http->offset > 0 &&
			       "http->offset unset while in trailer");
			stash_advance_rbuf(http, buf, buf_len);
			http->skip_rbuf_defer = 1;
			return;
		case MOG_CHUNK_STATE_DONE:
			commit_oneshot(http);
			assert(http->offset > 0 &&
			       "http->offset unset after chunk body done");
			stash_advance_rbuf(http, buf, buf_len);
			http->skip_rbuf_defer = 1;
		}
	}
}

static void
identity_body_after_header(struct mog_http *http, char *buf, size_t buf_len)
{
	size_t body_len = buf_len - http->offset;
	char *body_ptr = buf + http->offset;

	if (http->content_len < body_len)
		body_len = http->content_len;
	if (body_len == 0)
		return;
	http->offset += body_len;
	if (!mog_http_write_full(http->forward, body_ptr, body_len))
		stop(http, "500 Internal Server Error");
}

static bool lengths_ok(struct mog_http *http)
{
	if (http->content_len < 0)
		return false;	/* ERANGE */

	if (http->has_content_range) {
		if (http->chunked)
			return false;

		if (http->range_end < 0 || http->range_beg < 0)
			return false;	/* ERANGE */

		assert(http->range_end >= 0 && http->range_beg >= 0 &&
		       "bad range, http_parser.rl broken");

		/* can't end after we start */
		if (http->range_end < http->range_beg)
			return false;

		/*
		 * Content-Length should match Content-Range boundaries
		 * WARNING: Eric Wong sucks at arithmetic, check this:
		 */
		if (http->content_len >= 0) {
			off_t expect = http->range_end - http->range_beg + 1;

			if (http->content_len != expect)
				return false;
		}
	}
	return true;
}

void mog_http_put(struct mog_http *http, char *buf, size_t buf_len)
{
	char *path = mog_http_path(http, buf);
	struct mog_file *file;
	int open_flags = http->has_content_range ? O_CREAT : O_TRUNC | O_CREAT;

	if (!path)
		goto forbidden;	/* path traversal attack */
	assert(http->forward == NULL && "already have http->forward");
	assert(path[0] == '/' && "bad path");

	if (!lengths_ok(http)) {
		mog_http_resp(http, "400 Bad Request", false);
		return;
	}

	http->forward = mog_file_open_put(http->svc, path, open_flags);
	if (http->forward == NULL)
		goto err;

	file = &http->forward->as.file;
	assert(file->foff == 0 && "file->foff should be zero");
	if (http->has_content_range)
		file->foff = http->range_beg;
	if (http->has_trailer_md5 || http->has_expect_md5)
		file->md5 = mog_md5_new();

	if (buf_len == http->offset) {
		/* we got the HTTP header in one read() */
		if (http->chunked) {
			mog_rbuf_free_and_null(&http->rbuf);
			mog_chunk_init(http);
			http->offset = buf_len;
		}
		return;
	}
	/*
	 * otherwise we got part of the request body with the header,
	 * write partially read body
	 */
	assert(buf_len > http->offset && http->offset > 0
	       && "http->offset is wrong");

	if (http->chunked)
		chunked_body_after_header(http, buf, buf_len);
	else
		identity_body_after_header(http, buf, buf_len);

	return;
err:
	switch (errno) {
	case EACCES:
forbidden:
		mog_http_resp(http, "403 Forbidden", false);
		return;
	}

	mog_http_resp(http, "500 Internal Server Error", false);
}

static enum mog_next identity_put_in_progress(struct mog_fd *mfd)
{
	struct mog_http *http = &mfd->as.http;
	ssize_t r;
	size_t buf_len;
	char *buf;
	off_t need;

	assert(http->wbuf == NULL && "can't receive file with http->wbuf");
	assert(http->forward && http->forward != MOG_IOSTAT && "bad forward");

	need = http->content_len - http->forward->as.file.foff;
	if (http->has_content_range)
		need += http->range_beg;
	if (need == 0)
		return http_put_commit(http);

	assert(need > 0 && "over-wrote on PUT request");
	buf = mog_fsbuf_get(&buf_len);
	if (need < buf_len)
		buf_len = need;
retry:
	r = read(mfd->fd, buf, buf_len);
	if (r > 0) {
		if (!mog_http_write_full(http->forward, buf, r))
			return stop(http, "500 Internal Server Error");
		return MOG_NEXT_ACTIVE;
	}
	if (r != 0) {
		switch (errno) {
		case_EAGAIN: return MOG_NEXT_WAIT_RD;
		case EINTR: goto retry;
		}
	}
	/* assume all read() errors mean socket is unwritable, too */
	return stop(http, NULL);
}

static enum mog_next chunked_put_in_progress(struct mog_fd *mfd)
{
	struct mog_rbuf *rbuf;
	struct mog_http *http = &mfd->as.http;
	ssize_t r;
	size_t buf_len;
	size_t prev_len;
	char *buf;
	bool in_trailer = false;

	assert(http->wbuf == NULL && "can't receive file with http->wbuf");
	assert(http->forward && http->forward != MOG_IOSTAT && "bad forward");

	switch (http->chunk_state) {
	case MOG_CHUNK_STATE_DATA:
		assert(http->rbuf == NULL && "unexpected http->rbuf");
		if (http->content_len == 0) { /* final chunk */
			http->chunk_state = MOG_CHUNK_STATE_TRAILER;
			http->offset = 0;
			goto chunk_state_trailer;
		}
		assert(http->content_len > 0 && "bad chunk length");
		/* read the chunk into memory */
		buf = mog_fsbuf_get(&buf_len);
		if (buf_len > http->content_len)
			buf_len = http->content_len;
		do {
			r = read(mfd->fd, buf, buf_len);
		} while (r < 0 && errno == EINTR);

		if (r <= 0)
			goto read_err;
		if (!mog_http_write_full(http->forward, buf, r))
			return stop(http, "500 Internal Server Error");

		http->content_len -= r;

		/* chunk is complete */
		if (http->content_len == 0)
			mog_chunk_init(http);
		return MOG_NEXT_ACTIVE;
	case MOG_CHUNK_STATE_TRAILER:
chunk_state_trailer:
		in_trailer = true;
		/* fall-through */
	case MOG_CHUNK_STATE_SIZE:
		rbuf = http->rbuf;
		if (rbuf) {
			prev_len = rbuf->rsize;
			buf_len = MOG_HTTP_RBUF_SIZE - prev_len;
			buf = rbuf->rptr + prev_len;
		} else {
			prev_len = 0;
			rbuf = mog_rbuf_get(MOG_HTTP_RBUF_SIZE);
			buf_len = rbuf->rsize;
			buf = rbuf->rptr;
		}
		do {
			r = read(mfd->fd, buf, buf_len);
		} while (r < 0 && errno == EINTR);
		if (r <= 0)
			goto read_err;

		buf = rbuf->rptr;
		buf_len = r + prev_len;

		switch (mog_chunk_parse(http, buf, buf_len)) {
		case MOG_PARSER_ERROR:
			return stop(http, "400 Bad Request");
		case MOG_PARSER_CONTINUE:
			assert(http->chunk_state != MOG_CHUNK_STATE_DONE);
		case MOG_PARSER_DONE:
			switch (http->chunk_state) {
			case MOG_CHUNK_STATE_SIZE:
				if (in_trailer)
					assert(0 && "bad chunk state: size");
				/* client is trickling chunk size :< */
				mog_rbuf_free_and_null(&http->rbuf);
				http->offset = 0;
				return MOG_NEXT_ACTIVE;
			case MOG_CHUNK_STATE_DATA:
				if (in_trailer)
					assert(0 && "bad chunk state: data");
				/* client is trickling final chunk/trailer */
				mog_rbuf_free_and_null(&http->rbuf);
				return MOG_NEXT_ACTIVE;
			case MOG_CHUNK_STATE_TRAILER:
				http->rbuf = mog_rbuf_defer(rbuf);
				rbuf->rsize = buf_len;
				return MOG_NEXT_ACTIVE;
			case MOG_CHUNK_STATE_DONE:
				stash_advance_rbuf(http, buf, buf_len);

				/* pipelined HTTP request after trailers! */
				if (http->rbuf)
					assert(http->rbuf->rsize > 0
					       && http->offset == 0
					       && "bad rbuf");
				return http_put_commit(http);
			}
		}
		assert(0 && "compiler bug?");
	case MOG_CHUNK_STATE_DONE:
		assert(0 && "invalid state");
	}

read_err:
	if (r < 0) {
		switch (errno) {
		case_EAGAIN: return MOG_NEXT_WAIT_RD;
		}
	}
	return stop(http, NULL);
}

enum mog_next mog_http_put_in_progress(struct mog_fd *mfd)
{
	if (mfd->as.http.chunked)
		return chunked_put_in_progress(mfd);

	return identity_put_in_progress(mfd);
}
