#!/usr/bin/env ruby
# -*- encoding: binary -*-
# Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
# License: GPLv3 or later (see COPYING for details)
require 'test/test_helper'
require 'digest/md5'
require 'net/http'
require 'stringio'
require 'io/wait'

class TestHTTPPut < Test::Unit::TestCase
  EPOCH = "Thu, 01 Jan 1970 00:00:00 GMT"

  def replace_dates!(header)
    rv = header.gsub!(/: \w{3}, \d{2} \w{3} \d{4} \d{2}:\d{2}:\d{2} GMT\r\n/,
                      ": #{EPOCH}\r\n")
    assert(rv, header)
  end

  def setup
    @tmpdir = Dir.mktmpdir('cmogstored-httpput-test')
    @to_close = []
    @host = TEST_HOST
    srv = TCPServer.new(@host, 0)
    @port = srv.addr[1]
    srv.close
    @err = Tempfile.new("stderr")
    cmd = [ "cmogstored", "--docroot=#@tmpdir", "--httplisten=#@host:#@port",
            "--maxconns=500" ]
    vg = ENV["VALGRIND"] and cmd = vg.split(/\s+/).concat(cmd)
    @pid = fork {
      $stderr.reopen(@err)
      @err.close
      exec(*cmd)
    }
    tries = 300
    begin
      sleep 0.05
      @client = TCPSocket.new(@host, @port)
      @to_close << @client
      break
    rescue
    end while (tries-=1) > 0
  end

  def teardown
    Process.kill(:QUIT, @pid) rescue nil
    _, status = Process.waitpid2(@pid)
    @to_close.each { |io| io.close unless io.closed? }
    FileUtils.rm_rf(@tmpdir)
    @err.rewind
    $stderr.write(@err.read)
    assert status.success?, status.inspect
  end

  def test_put
    Net::HTTP.start(@host, @port) do |http|
      4.times do |i|
        put = Net::HTTP::Put.new("/foo#{i}")
        body = StringIO.new("BODY!")
        put.content_type = "application/octet-stream"
        put.body_stream = body
        put.content_length = body.size
        resp = http.request(put)
        assert_equal 201, resp.code.to_i
        assert_equal "BODY!", IO.read("#@tmpdir/foo#{i}")
      end
    end
  end

  def test_single_write
    req = "PUT /zz HTTP/1.1\r\n" \
          "Host: #@host:#@port\r\n" \
          "Content-Length: 5\r\n" \
          "\r\n" \
          "abcde"
    @client.write(req)
    expect = "HTTP/1.1 201 Created\r\n" \
      "Status: 201 Created\r\n" \
      "Date: #{EPOCH}\r\n" \
      "Content-Length: 0\r\n" \
      "Content-Type: text/plain\r\n" \
      "Connection: keep-alive\r\n\r\n"
    resp = @client.readpartial(666)
    replace_dates!(resp)
    assert_equal expect, resp
    assert_equal "abcde", IO.read("#@tmpdir/zz")
  end

  def test_zero_byte
    req = "PUT /zero HTTP/1.1\r\n" \
          "Host: #@host:#@port\r\n" \
          "Content-Length: 0\r\n" \
          "\r\n"
    @client.write(req)
    expect = "HTTP/1.1 201 Created\r\n" \
      "Status: 201 Created\r\n" \
      "Date: #{EPOCH}\r\n" \
      "Content-Length: 0\r\n" \
      "Content-Type: text/plain\r\n" \
      "Connection: keep-alive\r\n\r\n"
    resp = @client.readpartial(666)
    replace_dates!(resp)
    assert_equal expect, resp
    assert_equal "", IO.read("#@tmpdir/zero")
    assert_nil IO.select([@client], nil, nil, 0.1)
  end

  def test_pipelined
    req = 2.times.map do |i|
      "PUT /#{i} HTTP/1.1\r\n" \
      "Host: #@host:#@port\r\n" \
      "Content-Length: 5\r\n" \
      "\r\n" \
      "abcd#{i}"
    end.join
    @client.write(req)
    expect = "HTTP/1.1 201 Created\r\n" \
      "Status: 201 Created\r\n" \
      "Date: #{EPOCH}\r\n" \
      "Content-Length: 0\r\n" \
      "Content-Type: text/plain\r\n" \
      "Connection: keep-alive\r\n\r\n"
    resp = @client.read(expect.size * 2)
    replace_dates!(resp)
    assert_equal(expect * 2, resp)
    assert_equal "abcd0", IO.read("#@tmpdir/0")
    assert_equal "abcd1", IO.read("#@tmpdir/1")
    assert_nil IO.select([@client], nil, nil, 0.1)
  end

  def test_range_put
    req = "PUT /foo HTTP/1.0\r\n" \
          "Content-Length: 1\r\n" \
          "Content-Range: bytes 5-5/*\r\n" \
          "\r\na"
    @client.write(req)
    resp = @client.readpartial(666)
    assert_match(%r{\AHTTP/1\.1 201 Created\r\n}, resp)
    File.open("#@tmpdir/foo") do |fp|
      assert_equal("\0\0\0\0\0a", fp.read(fp.stat.size))
    end
  end

  def test_range_put_bad_content_length
    req = "PUT /foo HTTP/1.0\r\n" \
          "Content-Length: 2\r\n" \
          "Content-Range: bytes 5-5/*\r\n" \
          "\r\na"
    @client.write(req)
    resp = @client.read
    assert_match(%r{\AHTTP/1\.1 400 Bad Request\r\n}, resp)
    assert( ! File.exist?("#@tmpdir/foo") )
  end

  def test_range_put_bogus_range
    req = "PUT /foo HTTP/1.0\r\n" \
          "Content-Length: 2\r\n" \
          "Content-Range: bytes 5-1/*\r\n" \
          "\r\na"
    @client.write(req)
    resp = @client.read
    assert_match(%r{\AHTTP/1\.1 400 Bad Request\r\n}, resp)
    assert( ! File.exist?("#@tmpdir/foo") )
  end

  def test_put_content_len_overflow
    max = 0xffffffff << 64
    req = "PUT /foo HTTP/1.0\r\n" \
          "Content-Length: #{max}\r\n" \
          "\r\n"
    @client.write(req)
    resp = @client.read
    assert_match(%r{\AHTTP/1\.1 400 Bad Request\r\n}, resp)
    assert( ! File.exist?("#@tmpdir/foo") )
  end

  def test_put_range_beg_overflow
    max = 0xffffffff << 64
    req = "PUT /foo HTTP/1.1\r\n" \
          "Transfer-Encoding: chunked\r\n" \
          "Content-Range: bytes #{max}-#{max + 1}/*\r\n" \
          "\r\n"
    @client.write(req)
    resp = @client.read
    assert_match(%r{\AHTTP/1\.1 400 Bad Request\r\n}, resp)
    assert( ! File.exist?("#@tmpdir/foo") )
  end

  def test_put_range_end_overflow
    max = 0xffffffff << 64
    req = "PUT /foo HTTP/1.1\r\n" \
          "Transfer-Encoding: chunked\r\n" \
          "Content-Range: bytes 0-#{max}/*\r\n" \
          "\r\n"
    @client.write(req)
    resp = @client.read
    assert_match(%r{\AHTTP/1\.1 400 Bad Request\r\n}, resp)
    assert( ! File.exist?("#@tmpdir/foo") )
  end

  def test_put_range_real
    expect = ""
    Net::HTTP.start(@host, @port) do |http|
      beg = 0
      4.times do |i|
        put = Net::HTTP::Put.new("/foo")
        body = StringIO.new("BODY#{i}\n")
        put.content_type = "application/octet-stream"
        put.body_stream = body
        put["Content-Range"] = "bytes #{beg}-#{beg+body.size-1}/*"
        put.content_length = body.size
        resp = http.request(put)
        assert_equal 201, resp.code.to_i
        expect << body.string
        beg += body.size
      end
    end

    assert_equal expect, IO.read("#@tmpdir/foo")
  end

  def test_put_bad_content_length
    req = "PUT /foo HTTP/1.1\r\n" \
          "Content-Length: 123a\r\n" \
          "\r\n"
    @client.write(req)
    resp = @client.read
    assert_match(%r{\AHTTP/1\.1 400 Bad Request\r\n}, resp)
    assert( ! File.exist?("#@tmpdir/foo") )
  end

  def test_put_one_way
    req = "PUT /foo HTTP/1.1\r\n" \
          "Host: #@host:#@port\r\n" \
          "Content-Length: 0\r\n" \
          "\r\n"
    ok = 0
    begin
      nr = @client.write_nonblock(req)
      if nr == req.size
        ok += 1
      end
    rescue Errno::EAGAIN
      break
    end while true

    created = 0
    while @client.wait(5)
      line = @client.gets
      if line =~ %r{\AHTTP/1\.1 201 Created}
        created += 1
      end
    end
    assert_equal ok, created
    @client.close
  end

  def test_content_md5_good
    req = "PUT /foo HTTP/1.1\r\n" \
          "Host: #@host:#@port\r\n" \
          "Content-Length: 5\r\n" \
          "Content-MD5: q1a02StAcTrMWviZhdS3hg==\r\n" \
          "\r\n" \
          "abcde"
    @client.write(req)
    line = @client.gets
    assert_match(%r{\AHTTP/1\.1 201 Created}, line)
  end

  def test_content_md5_bad
    req = "PUT /foo HTTP/1.1\r\n" \
          "Host: #@host:#@port\r\n" \
          "Content-Length: 5\r\n" \
          "Content-MD5: q1a02StAcTrMWviZhdS3hg==\r\n" \
          "\r\n" \
          "abcd!"
    @client.write(req)
    line = @client.gets
    assert_match(%r{\AHTTP/1\.1 400 Bad Request}, line)
  end
end
